----- MODULE 2ConsensusUsingQueue -----
EXTENDS Naturals, TLC, Sequences

NONE == CHOOSE n : n \notin {0, 1}

(* An example algorithm that implements binary consensus for 2 processes
   using a queue and 2 atomic registers. *)

(* --algorithm 2ConsensusUsingQueueAlg

	(* 1. Shared objects: a queue (represented by a sequence Q) and 2 atomic registers *)

	variables Q = << "winner", "loser" >>, 
		  R = [i \in 1..2 |-> NONE], 

	(* 2. Process-private variables to store the input and the result for each process
	      and check invariants *)

		  pvalues = [i \in 1..2 |-> NONE],
	          values = [i \in 1..2 |-> NONE];

	(* 3. The definition of a queue -- a shared object we use. We don't need to define
	      atomic registers (declaring them as variables is enough). Also, we don't need
	      the "enqueue" operation of a queue, for we don't use it in the algorithm. 
	      Note that a macro can return a value, i.e., write to the variables passed
	      as parameters, because it is in-lined at compile-time. *)

	macro DEQ(result)
	begin
		result := Head(Q);
		Q := Tail(Q);
	end macro

	(* 4. The consensus algorithm implemented using two atomic registers and a queue.
	      Note that all input values of the algorithm should be passed as arguments
	      to the procedure, so that the algorithm is self-contained. We can use
	      process-private variables defined in part 2. only to return values. *)

	procedure propose(pval)
	variables result;
	begin
		write:	R[self] := pval;
		deq:	DEQ(result);
		decide:	if result = "winner" then values[self] := pval;
			else values[self] := R[3 - self];
			end if; 
		ret:	return;
	end procedure

	(* 5. An optional code for testing the consensus algorithm. The algorithm should be
	      self-contained, so that invoking just "propose(0)" or "propose(1)", without
	      writing to any variables, does not make the algorithm return wrong results
	      (although invariants may be violated then) *)
	
	process Proc \in 1..2
	begin
		pval:	either pvalues[self] := 0 or pvalues[self] := 1 end either;
		prop:	call propose(pvalues[self]);
	end process
	
	end algorithm
*)

\* BEGIN TRANSLATION
\* Label pval of process Proc at line 57 col 25 changed to pval_
VARIABLES Q, R, pvalues, values, pc, stack, pval, result

vars == << Q, R, pvalues, values, pc, stack, pval, result >>

ProcSet == (1..2)

Init == (* Global variables *)
        /\ Q = << "winner", "loser" >>
        /\ R = [i \in 1..2 |-> NONE]
        /\ pvalues = [i \in 1..2 |-> NONE]
        /\ values = [i \in 1..2 |-> NONE]
        (* Procedure propose *)
        /\ pval = [ self \in ProcSet |-> {}]
        /\ result = [ self \in ProcSet |-> {}]
        /\ stack = [self \in ProcSet |-> << >>]
        /\ pc = [self \in ProcSet |-> CASE self \in 1..2 -> "pval_"]

write(self) == /\ pc[self] = "write"
               /\ R' = [R EXCEPT ![self] = pval[self]]
               /\ pc' = [pc EXCEPT ![self] = "deq"]
               /\ UNCHANGED << Q, pvalues, values, stack, pval, result >>

deq(self) == /\ pc[self] = "deq"
             /\ result' = [result EXCEPT ![self] = Head(Q)]
             /\ Q' = Tail(Q)
             /\ pc' = [pc EXCEPT ![self] = "decide"]
             /\ UNCHANGED << R, pvalues, values, stack, pval >>

decide(self) == /\ pc[self] = "decide"
                /\ IF result[self] = "winner"
                      THEN /\ values' = [values EXCEPT ![self] = pval[self]]
                      ELSE /\ values' = [values EXCEPT ![self] = R[3 - self]]
                /\ pc' = [pc EXCEPT ![self] = "ret"]
                /\ UNCHANGED << Q, R, pvalues, stack, pval, result >>

ret(self) == /\ pc[self] = "ret"
             /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
             /\ result' = [result EXCEPT ![self] = Head(stack[self]).result]
             /\ pval' = [pval EXCEPT ![self] = Head(stack[self]).pval]
             /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
             /\ UNCHANGED << Q, R, pvalues, values >>

propose(self) == write(self) \/ deq(self) \/ decide(self) \/ ret(self)

pval_(self) == /\ pc[self] = "pval_"
               /\ \/ /\ pvalues' = [pvalues EXCEPT ![self] = 0]
                  \/ /\ pvalues' = [pvalues EXCEPT ![self] = 1]
               /\ pc' = [pc EXCEPT ![self] = "prop"]
               /\ UNCHANGED << Q, R, values, stack, pval, result >>

prop(self) == /\ pc[self] = "prop"
              /\ /\ pval' = [pval EXCEPT ![self] = pvalues[self]]
                 /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "propose",
                                                          pc        |->  "Done",
                                                          result    |->  result[self],
                                                          pval      |->  pval[self] ] >>
                                                      \o stack[self]]
              /\ result' = [result EXCEPT ![self] = {}]
              /\ pc' = [pc EXCEPT ![self] = "write"]
              /\ UNCHANGED << Q, R, pvalues, values >>

Proc(self) == pval_(self) \/ prop(self)

Next == (\E self \in ProcSet: propose(self))
           \/ (\E self \in 1..2: Proc(self))
           \/ (* Disjunct to prevent deadlock on termination *)
              (\A self \in ProcSet: pc[self] = "Done" /\ UNCHANGED vars)

Spec == Init /\ [][Next]_vars

Termination == <>(\A self \in ProcSet: pc[self] = "Done")

\* END TRANSLATION

Agreement == \A i, j \in 1..2 : (values[i] = NONE) \/ (values[j] = NONE) \/ (values[i] = values[j])

Validity == \A i \in 1..2 : (values[i] # NONE) => ( \E j \in 1..2 : (values[i] = pvalues[j]) )

=====
