----- MODULE Safe2RegularRegister -----
EXTENDS Naturals, TLC, Sequences
CONSTANT N

(* The algorithm implements a regular MRSW register using a safe MRSW register. *) 

(* --algorithm Safe2RegularRegister

	(* 1. Shared objects: a safe MRSW register. *)

	variables S = 0, iswriting = FALSE, 

	(* 2. Process-private variables (each memory location accessed by
	      at most one process) for storing procedure return values and
	      helper values. *)

		  Rvalue = [i \in 2..N |-> 0], \* Used by the readers
		  oldValue = 0;  \* Used by the writer 

	(* 3. The definition of a safe register. Note that each macro is executed
	      in a single step and that return values can be passed by writing
	      to variables passed as arguments (only for macros). *)

	macro RW_INIT()
	begin
		if self = 1 then iswriting := TRUE; end if; 
	end macro

	macro READ(result)
	begin
		if iswriting = FALSE then result := S;
		else either result := 0 or result := 1 end either;
		end if;
	end macro

	macro WRITE(value)
	begin
		S := value;
		iswriting := FALSE;
	end macro

	(* 4. The algorithm implementing a regular register using a safe one. *)
	
	procedure read()
	begin
		rinv:	RW_INIT();
		rresp:	READ(Rvalue[self]);
		rret:	return;
	end procedure

	procedure write(value)
	begin
		check:	if oldValue # value then
		winv:		RW_INIT();
		wresp:		WRITE(value);
		save:		oldValue := value;
			end if;
		wret:	return;
	end procedure

	(* 5. Code for testing the algorithm. Note that the algorithm defined in part 4.
	      should work regardless of the testing code here. For example, executing
	      "write(5)" or "read()" should give the expected results. *)

	process Writer = 1
	begin
		wploop:		while TRUE do
		wpwrite:		either call write(0) or call write(1) end either;
				end while;
	end process;

	process Readers \in 2..N
	begin
		rploop:		while TRUE do
		rpread:			call read();
				end while;
	end process;
			
	end algorithm
*)

\* BEGIN TRANSLATION
VARIABLES S, iswriting, Rvalue, oldValue, pc, stack, value

vars == << S, iswriting, Rvalue, oldValue, pc, stack, value >>

ProcSet == {1} \cup (2..N)

Init == (* Global variables *)
        /\ S = 0
        /\ iswriting = FALSE
        /\ Rvalue = [i \in 2..N |-> 0]
        /\ oldValue = 0
        (* Procedure write *)
        /\ value = [ self \in ProcSet |-> {}]
        /\ stack = [self \in ProcSet |-> << >>]
        /\ pc = [self \in ProcSet |-> CASE self = 1 -> "wploop"
                                        [] self \in 2..N -> "rploop"]

rinv(self) == /\ pc[self] = "rinv"
              /\ IF self = 1
                    THEN /\ iswriting' = TRUE
                    ELSE /\ TRUE
                         /\ UNCHANGED iswriting
              /\ pc' = [pc EXCEPT ![self] = "rresp"]
              /\ UNCHANGED << S, Rvalue, oldValue, stack, value >>

rresp(self) == /\ pc[self] = "rresp"
               /\ IF iswriting = FALSE
                     THEN /\ Rvalue' = [Rvalue EXCEPT ![self] = S]
                     ELSE /\ \/ /\ Rvalue' = [Rvalue EXCEPT ![self] = 0]
                             \/ /\ Rvalue' = [Rvalue EXCEPT ![self] = 1]
               /\ pc' = [pc EXCEPT ![self] = "rret"]
               /\ UNCHANGED << S, iswriting, oldValue, stack, value >>

rret(self) == /\ pc[self] = "rret"
              /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
              /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
              /\ UNCHANGED << S, iswriting, Rvalue, oldValue, value >>

read(self) == rinv(self) \/ rresp(self) \/ rret(self)

check(self) == /\ pc[self] = "check"
               /\ IF oldValue # value[self]
                     THEN /\ pc' = [pc EXCEPT 
                                      ![self] = "winv"]
                     ELSE /\ pc' = [pc EXCEPT 
                                      ![self] = "wret"]
               /\ UNCHANGED << S, iswriting, Rvalue, oldValue, stack, value >>

winv(self) == /\ pc[self] = "winv"
              /\ IF self = 1
                    THEN /\ iswriting' = TRUE
                    ELSE /\ TRUE
                         /\ UNCHANGED iswriting
              /\ pc' = [pc EXCEPT ![self] = "wresp"]
              /\ UNCHANGED << S, Rvalue, oldValue, stack, value >>

wresp(self) == /\ pc[self] = "wresp"
               /\ S' = value[self]
               /\ iswriting' = FALSE
               /\ pc' = [pc EXCEPT ![self] = "save"]
               /\ UNCHANGED << Rvalue, oldValue, stack, value >>

save(self) == /\ pc[self] = "save"
              /\ oldValue' = value[self]
              /\ pc' = [pc EXCEPT ![self] = "wret"]
              /\ UNCHANGED << S, iswriting, Rvalue, stack, value >>

wret(self) == /\ pc[self] = "wret"
              /\ pc' = [pc EXCEPT ![self] = Head(stack[self]).pc]
              /\ value' = [value EXCEPT ![self] = Head(stack[self]).value]
              /\ stack' = [stack EXCEPT ![self] = Tail(stack[self])]
              /\ UNCHANGED << S, iswriting, Rvalue, oldValue >>

write(self) == check(self) \/ winv(self) \/ wresp(self) \/ save(self)
                  \/ wret(self)

wploop == /\ pc[1] = "wploop"
          /\ pc' = [pc EXCEPT ![1] = "wpwrite"]
          /\ UNCHANGED << S, iswriting, Rvalue, oldValue, stack, value >>

wpwrite == /\ pc[1] = "wpwrite"
           /\ \/ /\ /\ stack' = [stack EXCEPT ![1] = << [ procedure |->  "write",
                                                          pc        |->  "wploop",
                                                          value     |->  value[1] ] >>
                                                      \o stack[1]]
                    /\ value' = [value EXCEPT ![1] = 0]
                 /\ pc' = [pc EXCEPT ![1] = "check"]
              \/ /\ /\ stack' = [stack EXCEPT ![1] = << [ procedure |->  "write",
                                                          pc        |->  "wploop",
                                                          value     |->  value[1] ] >>
                                                      \o stack[1]]
                    /\ value' = [value EXCEPT ![1] = 1]
                 /\ pc' = [pc EXCEPT ![1] = "check"]
           /\ UNCHANGED << S, iswriting, Rvalue, oldValue >>

Writer == wploop \/ wpwrite

rploop(self) == /\ pc[self] = "rploop"
                /\ pc' = [pc EXCEPT ![self] = "rpread"]
                /\ UNCHANGED << S, iswriting, Rvalue, oldValue, stack, value >>

rpread(self) == /\ pc[self] = "rpread"
                /\ stack' = [stack EXCEPT ![self] = << [ procedure |->  "read",
                                                         pc        |->  "rploop" ] >>
                                                     \o stack[self]]
                /\ pc' = [pc EXCEPT ![self] = "rinv"]
                /\ UNCHANGED << S, iswriting, Rvalue, oldValue, value >>

Readers(self) == rploop(self) \/ rpread(self)

Next == Writer
           \/ (\E self \in ProcSet: read(self) \/ write(self))
           \/ (\E self \in 2..N: Readers(self))
           \/ (* Disjunct to prevent deadlock on termination *)
              (\A self \in ProcSet: pc[self] = "Done" /\ UNCHANGED vars)

Spec == Init /\ [][Next]_vars

Termination == <>(\A self \in ProcSet: pc[self] = "Done")

\* END TRANSLATION

=====
